/**
 * Created by bjoern on 20.03.17.
 */

"use strict";

const request = require('request');

class treeCrawler {

    constructor(initialUrl) {

        if (initialUrl) this.initialUrl = initialUrl;

        this.config = {
            maxDepth: Number.POSITIVE_INFINITY,
            maxRequests: 2,
            maxThreads: 1,
            maxRetries: 3,
            sleepBetweenRequestMs: 200,
            urlScraper: treeCrawler.builtin.urlScraper.dummy,
            dataScraper: treeCrawler.builtin.dataScraper.dummy,
            log: console.warn,
            logEveryRequest: true,
            out: console.log,
        };
    }


    buildTree(callback) {
        const config = this.config;
        let l = config.log;

        {
            if (!this.initialUrl) throw new Error("initialUrl not set");
            if (config.maxThreads > 1) throw new Error("multithreading not implemented");
        }

        let visitedUrls = [];
        let foundUrls = [];
        let openJobs = [];
        let pendingRequests = 0; // this has no actual use beside finding errors regarding asynchronity

        let root = {children: [], url: this.initialUrl};
        foundUrls.push(this.initialUrl);

        openJobs.push({processNode: root, urlScraper: config.urlScraper, depth: 0});

        function onAllJobsDone() {
            callback(root);
        }

        //main processing loop
        function processNextJob() {
            if (openJobs.length < 1) {
                l("no more URLs left. ");
                onAllJobsDone();
                return;
            }

            if (visitedUrls.length >= config.maxRequests) {
                l("max requests exceeded. (" + config.maxRequests + ")");
                onAllJobsDone();
                return;
            }

            //do breadth first, or else a deep url will be marked as found, but not being visited because maxDepth, and later be skipped if found again on a higher level.
            let job = openJobs.shift();

            if (config.logEveryRequest)
                l("links in queue: " + openJobs.length + "\tcurrent depth: " + job.depth + " \t next url: " + job.processNode.url);

            setTimeout(traverseNode, config.sleepBetweenRequestMs, job.processNode, processNextJob, job.depth)
        }

        function traverseNode(node, callback, depth) {

            const depthToGo = config.maxDepth - depth;
            {
                //plausibility checks. None of these should ever trigger!
                if (depthToGo < 0) {
                    l("warning: depth was already reached.");
                    callback(node);
                    return;
                }

                if (visitedUrls.indexOf(node.url) > -1) {
                    l("warning url was already visited: " + node.url);
                    callback(node);
                    return;
                }

                if (visitedUrls.length > config.maxRequests) {
                    l("warning: max Urls were already exceeded. (" + config.maxRequests + ")");
                    callback(node);
                    return;
                }

                if (pendingRequests >= config.maxThreads) {
                    l("warning: too many concurrent requests.");
                    setTimeout(traverseNode, config.sleepBetweenRequestMs, node, config.urlScraper, callback, depth);
                    return;
                }
            }

            visitedUrls.push(node.url);

            pendingRequests++;
            requestDocument(node.url, config.maxRetries, function onResult(documentString) {
                pendingRequests--;

                node.children = [];
                {
                    let newChildren = [];
                    try {
                        newChildren = config.urlScraper(documentString, node.url);
                    } catch (e) {
                        l("error: urlScraper crashed");
                        l(e);
                    }

                    for (let i = 0; i < newChildren.length; i++) {
                        const newChild = newChildren[i];
                        let oldChild = node.children.find(x => newChild.url == x.url);
                        if (oldChild) {
                            oldChild.weight += newChild.weight;
                        } else {
                            node.children.push((newChild));
                        }
                    }
                }

                node.data = null;
                try {
                    node.data = config.dataScraper(documentString, node);
                } catch (e) {
                    l("error: dataScraper crashed");
                    l(e);
                }

                let unvisitedChildren = node.children.filter(x => foundUrls.indexOf(x.url) == -1);
                foundUrls = foundUrls.concat(unvisitedChildren.map(child => child.url));

                if (depthToGo > 1) {
                    for (let i = 0; i < unvisitedChildren.length; i++) {
                        openJobs.push({processNode: unvisitedChildren[i], depth: depth + 1});
                    }
                } else {
                    l("max depth reached.")
                }

                callback(node);
            })
        }

        function requestDocument(urlString, tryCount, callback) {

            request(urlString, function onResult(error, response, documentString) {
                if (error) {
                    l("error while parsing: " + urlString);
                    l(error);
                    if (tryCount > 0) {
                        l("retry...");
                        requestDocument(urlString, tryCount - 1, callback);
                    } else {
                        l("no more retry.");
                        callback();
                    }
                    return;
                }

                callback(documentString)
            })
        }

        processNextJob();
    }
}


treeCrawler.builtin = {
    dataScraper: require("./builtin/dataScraper"),
    urlScraper: require("./builtin/urlScraper"),
    treeHelper: require("./builtin/treeHelper")
};


module.exports = treeCrawler;